﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;
using System.Windows.Forms.DataVisualization.Charting;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Core.Views;
using HeuristicLab.Data;
using HeuristicLab.MainForm;

namespace HeuristicLab.Optimization.BubbleChart {
  [View("Bubble Chart (Recursive)")]
  [Content(typeof(RecursiveDataItem), false)]
  public partial class BubbleChartView : ItemView {
    private const string Separator = ":";

    private enum SizeDimension { Constant = 0 }
    private enum AxisDimension { Index = 0 }

    private readonly Dictionary<RecursiveDataItem, int> itemToIndexMapping = new Dictionary<RecursiveDataItem, int>();
    private readonly Dictionary<string, Dictionary<object, double>> categoricalMapping = new Dictionary<string, Dictionary<object, double>>();

    private string XAxisValue { get { return (string)xAxisComboBox.SelectedItem; } }
    private string YAxisValue { get { return (string)yAxisComboBox.SelectedItem; } }
    private string SizeAxisValue { get { return (string)sizeComboBox.SelectedItem; } }

    private bool updating = false;

    public new RecursiveDataItem Content {
      get { return (RecursiveDataItem)base.Content; }
      set { base.Content = value; }
    }

    public BubbleChartView() {
      InitializeComponent();

      chart.CustomizeAllChartAreas();
      chart.ChartAreas[0].CursorX.Interval = 1;
      chart.ChartAreas[0].CursorY.Interval = 1;
      chart.ChartAreas[0].AxisX.ScaleView.Zoomable = true;
      chart.ChartAreas[0].AxisY.ScaleView.Zoomable = true;
    }

    protected override void OnContentChanged() {
      base.OnContentChanged();

      UpdateTreeView();
      UpdateLevelControl();

      UpdateComboBoxes();
      UpdateDataPoints();
      //UpdateCaption();
    }
    protected override void SetEnabledStateOfControls() {
      base.SetEnabledStateOfControls();
      levelNumericUpDown.Enabled = Content != null;
      // ToDo
    }

    #region Tree Queries
    private IEnumerable<RecursiveDataItem> GetAvailableItems() {
      return IterateCheckedNodes()
        .Select(n => (RecursiveDataItem)n.Tag);
    }
    private IEnumerable<string> GetAvailableKeys() {
      var collector = new HashSet<string>();
      GetAvailableKeys(Content, collector);
      return collector;
    }
    private void GetAvailableKeys(RecursiveDataItem node, ISet<string> collector) {
      foreach (var child in node.Children) {
        foreach (var key in child.Data.Keys) {
          collector.Add(node.Name + Separator + key);
        }
      }

      foreach (var child in node.Children)
        GetAvailableKeys(child, collector);
    }
    #endregion

    #region Update Controls
    private void UpdateLevelControl() {
      if (Content == null) return;
      if (treeView.Nodes.Count > 0)
        levelNumericUpDown.Maximum = IterateAllNodes().Max(t => t.Level);
      else
        levelNumericUpDown.Maximum = 0;
    }

    private void UpdateTreeView() {
      treeView.Nodes.Clear();
      if (Content != null)
        treeView.Nodes.Add(CreateTreeNode(Content));
      treeView.ExpandAll();
      splitContainer.Panel1Collapsed = treeView.Nodes.Count == 0;
      if (treeView.Nodes.Count > 0)
        treeView.SelectedNode = treeView.Nodes[0];
    }
    private TreeNode CreateTreeNode(RecursiveDataItem item) {
      var node = new TreeNode(item.Name) {
        Tag = item,
        Checked = true
      };
      foreach (var child in item.Children)
        node.Nodes.Add(CreateTreeNode(child));
      return node;
    }

    private void UpdateTreeViewCheckBoxes() {
      int level = (int)levelNumericUpDown.Value;
      bool includeChildren = includeChildrenCheckBox.Checked;
      foreach (var node in IterateAllNodes()) {
        bool @checked = includeChildren ? node.Level >= level : node.Level == level;
        if (node.Checked != @checked)
          node.Checked = @checked;
      }
    }

    private void UpdateComboBoxes() {
      var selectedXAxis = (string)xAxisComboBox.SelectedItem;
      var selectedYAxis = (string)yAxisComboBox.SelectedItem;
      var selectedSizeAxis = (string)sizeComboBox.SelectedItem;
      xAxisComboBox.Items.Clear();
      yAxisComboBox.Items.Clear();
      sizeComboBox.Items.Clear();

      if (Content != null) {
        var axisNames = GetAvailableKeys().ToArray();
        var additionalAxisDimension = Enum.GetNames(typeof(AxisDimension));
        var additionalSizeDimension = Enum.GetNames(typeof(SizeDimension));

        xAxisComboBox.Items.AddRange(additionalAxisDimension);
        xAxisComboBox.Items.AddRange(axisNames);
        yAxisComboBox.Items.AddRange(additionalAxisDimension);
        yAxisComboBox.Items.AddRange(axisNames);
        sizeComboBox.Items.AddRange(additionalSizeDimension);
        sizeComboBox.Items.AddRange(axisNames);

        bool changed = false;
        if (selectedXAxis != null && xAxisComboBox.Items.Contains(selectedXAxis)) {
          xAxisComboBox.SelectedItem = selectedXAxis;
          changed = true;
        }
        if (selectedYAxis != null && yAxisComboBox.Items.Contains(selectedYAxis)) {
          yAxisComboBox.SelectedItem = selectedYAxis;
          changed = true;
        }
        if (selectedSizeAxis != null && sizeComboBox.Items.Contains(selectedSizeAxis)) {
          sizeComboBox.SelectedItem = selectedSizeAxis;
          changed = true;
        } else sizeComboBox.SelectedItem = SizeDimension.Constant.ToString();
        if (changed) {
          UpdateDataPoints();
          UpdateAxisLabels();
        }
      }
    }

    private void UpdateDataPoints() {
      var series = chart.Series[0];
      series.Points.Clear();
      itemToIndexMapping.Clear();
      categoricalMapping.Clear();
      RebuildInverseIndex();

      chart.ChartAreas[0].AxisX.IsMarginVisible = !(XAxisValue != null && XAxisValue.Contains(AxisDimension.Index.ToString()));
      chart.ChartAreas[0].AxisY.IsMarginVisible = !(YAxisValue != null && YAxisValue.Contains(AxisDimension.Index.ToString()));

      if (Content != null) {
        var items = GetAvailableItems().ToList();
        var xss = GetValues(items, XAxisValue).Select(x => x.ToList()).ToList();
        var yss = GetValues(items, YAxisValue).Select(x => x.ToList()).ToList();
        var sss = GetValues(items, SizeAxisValue).Select(x => x.ToList()).ToList();

        if (xss.Any(xs => xs.Any()) && yss.Any(ys => ys.Any()) && sss.Any(ss => ss.Any())) {
          var xDim = Tuple.Create(xss.Count, xss.Max(xs => xs.Count));
          var yDim = Tuple.Create(yss.Count, yss.Max(ys => ys.Count));
          var sDim = Tuple.Create(sss.Count, sss.Max(ss => ss.Count));

          var isSimple = new Func<Tuple<int, int>, bool>(dim => dim.Item1 == 1 && dim.Item2 > 1 || dim.Item1 > 1 && dim.Item2 == 1);
          var flatten = new Func<List<List<double>>, List<double>>(ss => ss.SelectMany(s => s).ToList());

          if (isSimple(xDim) && isSimple(yDim)) {
            var xs = flatten(xss);
            var ys = flatten(yss);
            if (xs.Count == ys.Count) {
              var xsEnum = xs.GetEnumerator();
              var ysEnum = ys.GetEnumerator();
              while (xsEnum.MoveNext() & ysEnum.MoveNext())
                series.Points.Add(new DataPoint(xsEnum.Current, new[] { ysEnum.Current, sss[0][0] }));
            } /*else {
              foreach (var x in xs)
                foreach (var y in ys)
                  series.Points.Add(new DataPoint(x, new[] { y, sss[0][0] }));
            }*/
          } else if (isSimple(xDim)) {
            var xs = flatten(xss);
            if (xs.Count == yss.Count) {
              var xsEnum = xs.GetEnumerator();
              var yssEnum = yss.GetEnumerator();
              while (xsEnum.MoveNext() & yssEnum.MoveNext()) {
                var x = xsEnum.Current;
                var ys = yssEnum.Current;
                foreach (var y in ys)
                  series.Points.Add(new DataPoint(x, new[] { y, sss[0][0] }));
              }
            }
          } else if (isSimple(yDim)) {
            var ys = flatten(yss);
            if (xss.Count == ys.Count) {
              var xssEnum = xss.GetEnumerator();
              var ysEnum = ys.GetEnumerator();
              while (xssEnum.MoveNext() & ysEnum.MoveNext()) {
                var xs = xssEnum.Current;
                var y = ysEnum.Current;
                foreach (var x in xs)
                  series.Points.Add(new DataPoint(x, new[] { y, sss[0][0] }));
              }
            }
          } else {
            if (xss.Count == yss.Count) {
              var xssEnum = xss.GetEnumerator();
              var yssEnum = yss.GetEnumerator();
              while (xssEnum.MoveNext() & yssEnum.MoveNext()) {
                var xs = xssEnum.Current;
                var ys = yssEnum.Current;
                if (xs.Count == ys.Count) {
                  var xsEnum = xs.GetEnumerator();
                  var ysEnum = ys.GetEnumerator();
                  while (xsEnum.MoveNext() & ysEnum.MoveNext())
                    series.Points.Add(new DataPoint(xsEnum.Current, new[] { ysEnum.Current, sss[0][0] }));
                }
              }
            } else if (xDim.Item2 == yDim.Item2) {
              foreach (var xs in xss) {
                foreach (var ys in yss) {
                  var xsEnum = xs.GetEnumerator();
                  var ysEnum = ys.GetEnumerator();
                  while (xsEnum.MoveNext() & ysEnum.MoveNext()) {
                    series.Points.Add(new DataPoint(xsEnum.Current, new[] { ysEnum.Current, sss[0][0] }));
                  }
                }
              }
            }
          }
        }
        UpdateMarkerSizes();
      }
    }

    /*var items = GetAvailableItems();
          foreach (var item in items) {
            var xs = GetValues(item, XAxisValue).ToList();
            var ys = GetValues(item, YAxisValue).ToList();
            var ss = GetValues(item, SizeAxisValue).ToList();
            if (xs.Count > 0 && ys.Count > 0 && (ss.Count == 1 || ss.Count == xs.Count || ss.Count == ys.Count)) {
              var sEnum = ss.Count > 1 ? ss.GetEnumerator() : Enumerable.Repeat(ss[0], xs.Count * ys.Count).GetEnumerator();
              sEnum.MoveNext();
              if (xs.Count > 1 && ys.Count > 1) { // index matching
                for (int i = 0; i < Math.Min(xs.Count, ys.Count); i++)
                  series.Points.Add(new DataPoint(xs[i], new[] { ys[i], sEnum.Current }));
              } else {
                foreach (var x in xs) {
                  foreach (var y in ys) {
                    series.Points.Add(new DataPoint(x, new[] { y, sEnum.Current }));
                    sEnum.MoveNext();
                  }
                }
              }
            }
          }*/

    private IEnumerable<IEnumerable<double>> GetValues(IEnumerable<RecursiveDataItem> items, string key) {
      if (key == null) yield break;
      var keyTokens = key.Split(new[] { Separator }, StringSplitOptions.RemoveEmptyEntries);
      if (keyTokens.Length == 1) {
        if (Enum.IsDefined(typeof(SizeDimension), key)) {
          var sizeDimension = (SizeDimension)Enum.Parse(typeof(SizeDimension), key);
          yield return GetValue(null, sizeDimension).ToEnumerable();
        } else
          foreach (var item in items) {
            if (Enum.IsDefined(typeof(AxisDimension), key)) {
              var axisDimension = (AxisDimension)Enum.Parse(typeof(AxisDimension), key);
              yield return GetValue(item, axisDimension).ToEnumerable();
            } else if (item.Data.ContainsKey(key)) {
              yield return item.Data[key] as IEnumerable<double> ?? ConvertToDouble(item, key).ToEnumerable();
            }
          }
      } else if (keyTokens.Length == 2) {
        string parentName = keyTokens[0];
        key = keyTokens[1];
        foreach (var item in items.Where(item => item.Name == parentName)) {
          foreach (var child in item.Children.Where(child => child.Data.ContainsKey(key))) {
            yield return child.Data[key] as IEnumerable<double> ?? ConvertToDouble(child, key).ToEnumerable();
          }
        }
      } else throw new InvalidOperationException("key either contains a single key or a child/data key");

    }

    private IEnumerable<double> GetValues(RecursiveDataItem parent, string key) {
      if (parent == null || string.IsNullOrEmpty(key))
        return Enumerable.Empty<double>();
      var tokens = key.Split(new[] { Separator }, StringSplitOptions.RemoveEmptyEntries);
      if (tokens.Length == 2) {
        var parentKey = tokens[0];
        var dataKey = tokens[1];
        if (parent.Name == parentKey) {
          return parent.Children
            .Where(child => child.Data.ContainsKey(dataKey))
            .Select(child => ConvertToDouble(child, dataKey));
        }
        return Enumerable.Empty<double>();
      } else {
        var item = parent;
        if (item == null || string.IsNullOrEmpty(key))
          return Enumerable.Empty<double>();

        if (Enum.IsDefined(typeof(AxisDimension), key)) {
          var axisDimension = (AxisDimension)Enum.Parse(typeof(AxisDimension), key);
          return GetValue(item, axisDimension).ToEnumerable();
        } else if (Enum.IsDefined(typeof(SizeDimension), key)) {
          var sizeDimension = (SizeDimension)Enum.Parse(typeof(SizeDimension), key);
          return GetValue(item, sizeDimension).ToEnumerable();
        } else if (item.Data.ContainsKey(key)) {

          return ConvertToDouble(item, key).ToEnumerable();
        } else {
          return Enumerable.Empty<double>();
        }
      }
    }

    private double ConvertToDouble(RecursiveDataItem item, string key) {
      IItem value = item.Data[key];
      var doubleValue = value as DoubleValue;
      var intValue = value as IntValue;
      var timeSpanValue = value as TimeSpanValue;
      double? ret = null;
      if (doubleValue != null) {
        if (!double.IsNaN(doubleValue.Value) && !double.IsInfinity(doubleValue.Value))
          ret = doubleValue.Value;
      } else if (intValue != null)
        ret = intValue.Value;
      else if (timeSpanValue != null)
        ret = timeSpanValue.Value.TotalSeconds;
      else
        ret = GetCategoricalValue(key, value.ToString());
      return ret.Value;
    }

    private double? GetCategoricalValue(string key, string value) {
      if (!categoricalMapping.ContainsKey(key)) {
        categoricalMapping[key] = new Dictionary<object, double>();
        var orderedCategories =
          GetAvailableItems().Where(x => x.Data.ContainsKey(key)).Select(x => x.Data[key].ToString())
            .Distinct().OrderBy(x => x, new NaturalStringComparer());
        int count = 1;
        foreach (var category in orderedCategories) {
          categoricalMapping[key].Add(category, count);
          count++;
        }
      }
      if (!this.categoricalMapping[key].ContainsKey(value)) return null;
      return this.categoricalMapping[key][value];
    }
    private double GetValue(RecursiveDataItem item, AxisDimension axisDimension) {
      double value = double.NaN;
      switch (axisDimension) {
        case AxisDimension.Index:
          value = itemToIndexMapping[item];
          break;
        default:
          throw new ArgumentException("No handling strategy for " + axisDimension.ToString() + " is defined.");
      }
      return value;
    }
    private double GetValue(RecursiveDataItem item, SizeDimension sizeDimension) {
      double value = double.NaN;
      switch (sizeDimension) {
        case SizeDimension.Constant:
          value = 5;
          break;
        default:
          throw new ArgumentException("No handling strategy for " + sizeDimension.ToString() + " is defined.");
      }
      return value;
    }
    private void RebuildInverseIndex() {
      if (Content != null) {
        itemToIndexMapping.Clear();
        int i = 0;
        foreach (var item in GetAvailableItems()) {
          // ToDo: do not add if key (which one?) is not present within item
          itemToIndexMapping.Add(item, i);
          i++;
        }
      }
    }

    private void UpdateAxisLabels() {
      Axis xAxis = chart.ChartAreas[0].AxisX;
      Axis yAxis = chart.ChartAreas[0].AxisY;
      //mkommend: combobox.SelectedIndex could not be used as this changes during hovering over possible values
      var xSAxisSelected = XAxisValue == null ? null : (string)xAxisComboBox.SelectedItem;
      var ySAxisSelected = YAxisValue == null ? null : (string)yAxisComboBox.SelectedItem;
      SetCustomAxisLabels(xAxis, xSAxisSelected);
      SetCustomAxisLabels(yAxis, ySAxisSelected);
      if (XAxisValue != null)
        xAxis.Title = XAxisValue;
      if (YAxisValue != null)
        yAxis.Title = YAxisValue;
    }
    private void chart_AxisViewChanged(object sender, System.Windows.Forms.DataVisualization.Charting.ViewEventArgs e) {
      this.UpdateAxisLabels();
    }
    private void SetCustomAxisLabels(Axis axis, string key) {
      if (key == null) return;
      var tokens = key.Split(new[] { Separator }, StringSplitOptions.RemoveEmptyEntries);
      if (tokens.Length == 2)
        key = tokens[1];
      axis.CustomLabels.Clear();
      if (categoricalMapping.ContainsKey(key)) {
        foreach (var pair in categoricalMapping[key]) {
          string labelText = pair.Key.ToString();
          CustomLabel label = new CustomLabel();
          label.ToolTip = labelText;
          if (labelText.Length > 25)
            labelText = labelText.Substring(0, 25) + " ... ";
          label.Text = labelText;
          label.GridTicks = GridTickTypes.TickMark;
          label.FromPosition = pair.Value - 0.5;
          label.ToPosition = pair.Value + 0.5;
          axis.CustomLabels.Add(label);
        }
      } else if (Content.Data.ContainsKey(key) && Content.Data[key] is TimeSpanValue) { // TODO
        chart.ChartAreas[0].RecalculateAxesScale();
        for (double i = axis.Minimum; i <= axis.Maximum; i += axis.LabelStyle.Interval) {
          TimeSpan time = TimeSpan.FromSeconds(i);
          string x = string.Format("{0:00}:{1:00}:{2:00}", time.Hours, time.Minutes, time.Seconds);
          axis.CustomLabels.Add(i - axis.LabelStyle.Interval / 2, i + axis.LabelStyle.Interval / 2, x);
        }
      }
    }

    private void UpdateMarkerSizes() {
      var series = chart.Series[0];
      if (series.Points.Count <= 0) return;

      var sizeValues = series.Points.Select(p => p.YValues[1]);
      double minSizeValue = sizeValues.Min();
      double maxSizeValue = sizeValues.Max();
      double sizeRange = maxSizeValue - minSizeValue;

      const int smallestBubbleSize = 5;

      foreach (DataPoint point in series.Points) {
        //calculates the relative size of the data point  0 <= relativeSize <= 1
        double relativeSize = (point.YValues[1] - minSizeValue);
        if (sizeRange > double.Epsilon) {
          relativeSize /= sizeRange;

          //invert bubble sizes if the value of the trackbar is negative
          if (sizeTrackBar.Value < 0) relativeSize = Math.Abs(relativeSize - 1);
        } else relativeSize = 1;

        double sizeChange = Math.Abs(sizeTrackBar.Value) * relativeSize;
        point.MarkerSize = (int)Math.Round(sizeChange + smallestBubbleSize);
      }
    }
    #endregion

    #region Event Handlers
    private void treeView_AfterCheck(object sender, TreeViewEventArgs e) {
      if (updating) return;
      updating = true;
      UpdateComboBoxes();
      UpdateDataPoints();
      updating = false;
    }
    private void treeView_AfterSelect(object sender, TreeViewEventArgs e) {
      //UpdateNodeColors();
      //UpdateComboBoxes();
      //UpdateDataPoints();
    }

    private void axisComboBox_SelectedValueChanged(object sender, EventArgs e) {
      if (updating) return;
      updating = true;
      UpdateDataPoints();
      UpdateAxisLabels();
      updating = false;
    }

    private void includeChildrenCheckBox_CheckedChanged(object sender, EventArgs e) {
      if (updating) return;
      updating = true;
      UpdateTreeViewCheckBoxes();
      UpdateComboBoxes();
      updating = false;
    }

    private void levelNumericUpDown_ValueChanged(object sender, EventArgs e) {
      if (updating) return;
      updating = true;
      UpdateTreeViewCheckBoxes();
      UpdateComboBoxes();
      updating = false;
    }

    private void sizeTrackBar_ValueChanged(object sender, EventArgs e) {
      UpdateMarkerSizes();
    }
    #endregion

    #region Helpers
    private IEnumerable<TreeNode> IterateAllNodes() {
      return Collect(treeView.Nodes);
    }
    private IEnumerable<TreeNode> IterateCheckedNodes() {
      return IterateAllNodes().Where(t => t.Checked);
    }
    private IEnumerable<TreeNode> Collect(TreeNodeCollection nodes) {
      foreach (TreeNode node in nodes) {
        yield return node;
        foreach (var child in Collect(node.Nodes))
          yield return child;
      }
    }
    #endregion
  }
}
